@@ -24,7 +24,7 @@ import errno
 
 from cgutils import host
 from cgutils import process
-from . import fileops
+from cgutils import fileops
 
 
 if sys.version_info.major == 3:
@@ -207,6 +207,26 @@ class PercpuStat(dict):
         return ret
 
 
+class CpuacctUsageAllStat(dict):
+    @staticmethod
+    def parse(content):
+        # cpu user system
+        # 0 26703258962 1612927964
+        # 1 37879159293 1963003474
+
+        ret = {}
+        lines = content.split('\n')
+        header = lines.pop(0)
+        if header != "cpu user system":
+            raise EnvironmentError("Invalid header: %s" % header)
+        lines.remove('')
+
+        for line in lines:
+            cpuid, user, system = line.split(' ')
+            ret[int(cpuid)] = (long(user), long(system))
+        return ret
+
+
 class SlabinfoStat(dict):
     def __parse_version_2_1(lines):
         # Drop legend '# name            <active_objs> <num_objs> <objsize> <objperslab> <pagesperslab> : tunables <limit> <batchcount> <sharedfactor> : slabdata <active_slabs> <num_slabs> <sharedavail>'
@@ -270,7 +290,49 @@ class SlabinfoStat(dict):
         if m:
             return SlabinfoStat._PARSERS[m.group(1)](lines)
         else:
-            raise EnvironmentError('No header found in memory.kmem.slabinfo')
+            return None
+
+
+class PidsEventsStat(dict):
+    @staticmethod
+    def parse(content):
+        ret = {}
+        lines = content.split('\n')
+        lines.remove('')
+        for line in lines:
+            name, val = line.split(' ')
+            ret[name] = long(val)
+        return ret
+
+
+class RdmaStat(dict):
+    __RE = '^(?P<name>\w+)\s+hca_handle=(?P<hca_handle>\d+)\s+hca_object=(?P<hca_object>(\d+|max))'
+    _RE = re.compile(__RE)
+
+    @staticmethod
+    def parse(content):
+        """ Parse rdma.curren and rdma.max
+
+        Example contents:
+          mlx4_0 hca_handle=2 hca_object=2000
+          ocrdma1 hca_handle=3 hca_object=max
+
+        >>> RdmaStat.parse("mlx4_0 hca_handle=2 hca_object=2000\\nocrdma1 hca_handle=3 hca_object=max")
+        {'mlx4_0': {'hca_handle': 2, 'hca_object': 2000}, 'ocrdma1': {'hca_handle': 3, 'hca_object': 'max'}}
+        """
+        ret = {}
+        lines = content.split('\n')
+        for line in lines:
+            m = RdmaStat._RE.match(line)
+            if m is None:
+                continue
+            name = m.group('name')
+            hca_handle = long(m.group('hca_handle'))
+            hca_object = m.group('hca_object')
+            if hca_object != "max":
+                hca_object = long(hca_object)
+            ret[name] = {"hca_handle": hca_handle, "hca_object": hca_object}
+        return ret
 
 
 #
@@ -314,7 +376,12 @@ class SubsystemCpuacct(Subsystem):
     STATS = {
         'usage': long,
         'stat': SimpleStat,
-        'usage_percpu': PercpuStat,
+        'usage_all': CpuacctUsageAllStat,
+        'usage_sys':  long,
+        'usage_user': long,
+        'usage_percpu':      PercpuStat,
+        'usage_percpu_sys':  PercpuStat,
+        'usage_percpu_user': PercpuStat,
     }
 
 
@@ -405,6 +472,8 @@ class SubsystemBlkio(Subsystem):
         'sectors_recursive': SimpleStat,
         'throttle.io_service_bytes': BlkioStat,
         'throttle.io_serviced': BlkioStat,
+        'throttle.io_service_bytes_recursive': BlkioStat,
+        'throttle.io_serviced_recursive': BlkioStat,
         'time': SimpleStat,
         'time_recursive': SimpleStat,
         # Debugging files (Appeared only CONFIG_DEBUG_BLK_CGROUP=y)
@@ -473,14 +542,39 @@ class SubsystemHugetlb(Subsystem):
     MAX_ULONGLONG = 2 ** 63 - 1
     # FIXME: this is x86 specific and we have to treat
     # different size of HugeTLB on other architectures.
-    __SIZE = '2MB'
+    __stats = {}
+    __configs = {}
+    for __size in ('2MB', '1GB'):
+        __stats.update({
+            __size + '.failcnt': long,
+            __size + '.max_usage_in_bytes': long,
+            __size + '.usage_in_bytes': long,
+        })
+        __configs.update({
+            __size + '.limit_in_bytes': MAX_ULONGLONG,
+        })
+    STATS = __stats
+    CONFIGS = __configs
+
+
+class SubsystemPids(Subsystem):
+    NAME = 'pids'
     STATS = {
-        __SIZE + '.failcnt': long,
-        __SIZE + '.max_usage_in_bytes': long,
-        __SIZE + '.usage_in_bytes': long,
+        'current': long,
+        'events': PidsEventsStat,
     }
     CONFIGS = {
-        __SIZE + '.limit_in_bytes': MAX_ULONGLONG,
+        'max': 'max',
+    }
+
+
+class SubsystemRdma(Subsystem):
+    NAME = 'rdma'
+    STATS = {
+        'current': RdmaStat,
+    }
+    CONFIGS = {
+        'max': RdmaStat,
     }
 
 
@@ -503,6 +597,8 @@ _subsystem_name2class = {
     'devices': SubsystemDevices,
     'net_prio': SubsystemNetPrio,
     'hugetlb': SubsystemHugetlb,
+    'pids': SubsystemPids,
+    'rdma': SubsystemRdma,
 }
 
 
@@ -560,6 +656,9 @@ class CGroup:
         NumaStat: NumaStat.parse,
         PercpuStat: PercpuStat.parse,
         SlabinfoStat: SlabinfoStat.parse,
+        CpuacctUsageAllStat: CpuacctUsageAllStat.parse,
+        PidsEventsStat: PidsEventsStat.parse,
+        RdmaStat: RdmaStat.parse,
     }
 
     def _calc_depth(self, path):
